/*

MedImage Wound Analysis Add-on - Hue grabber
==============================

Developed by (C) AtomJump Ltd. 10 Aug 2017. This is a commercial add-on to MedImage, and is licensed
by AtomJump on a per-server basis.

This script accepts an input .json file that defines how it will be run.

The json file includes a photo filename, and an x,y coordinate within that photo. 
It also gives the choice of
  - low end of HSV colour range, or upper end, or use an auto range around the selected colour
  - wound colours or sticker colours
  - master or disciple json
  
The script will read the photo file, grab the RGB of that coordinate, convert to HSV and write 
the value into a new colour range object in either the config/master.json, or the photo's json file.

e.g. but now as parameters in a get request 
{
    "photo": "/var/www/html/medimageservtest/photos/yo/9-Aug-2017-15-17-45.jpg",
    "x": 1386,
    "y": 1955,
    "lowRange": true,
    "type": "wound",
    "master": false,
    "keep": true
}


Note:  x is from the right side of the photo (= 0) moving left and increasing.
       y is from the top side of the photo (=0) move down and increasing.


*/


var cv = require('opencv');
var fs = require('fs');
var fsExtra = require('fs-extra');
var upath = require('upath');
var path = require('path');
var queryString = require('querystring');
var exec = require('child_process').exec;
var resolve = require('path').resolve;

var verbose = false;

var AUTO_HUE_RANGE = 3;				//Range in hue (H in an HSV colour space)
var AUTO_SATURATION_RANGE = 50;		//Range in saturation (S in an HSV colour space)
var AUTO_VALUE_RANGE = 50;			//Range in brightness (V in an HSV colour space)



var masterConfigFile = __dirname + '/config/master.json';
var globalConfigFile = __dirname + '/../../config.json';
var relImagePath = "../../photos";
var mainMedImagePath = path.relative(process.cwd(), __dirname + '/' + relImagePath);


//Include shared funcs
var normalizeInclWinNetworks = require(__dirname + '/common/libfuncs.js').normalizeInclWinNetworks;
var readConfig = require(__dirname + '/common/libfuncs.js').readConfig;
var writeConfig = require(__dirname + '/common/libfuncs.js').writeConfig;




function checkRangeExist(colourRanges, newRange) {
	//returns false if the newRange doesn't exist, true if it does already
	var newRangeStr = JSON.stringify(newRange);
	for(var cnt=0; cnt< colourRanges.length; cnt++) {
		if(JSON.stringify(colourRanges[cnt]) === newRangeStr) {
			return true;
		}	
	}
	
	return false;
}



if(process.argv[2]) {

	//Incoming get requests are in normal "var=value&var2=value" format urlencoded
	var opts = queryString.parse(decodeURIComponent(process.argv[2]));
	var origPhoto = opts.photo;
	opts.photo = __dirname + "/" + relImagePath + "/" + opts.photo;		
		

	if(verbose == true) console.log("Options are:" + JSON.stringify(opts, null, 2));

	//Now read the photo file in
	cv.readImage(opts.photo, function(err, im) {
		if (err) throw err;
		if (im.width() < 1 || im.height() < 1) throw new Error('Error: The image has no size, or does not appear valid. Parameters input=\n' + JSON.stringify(opts, null, 2));
		
		
		if(opts.x) {
			//We need to scale these, which are between (0 - opts.width) and (0 - opts.height)
			//back to the scale of the image (given by im.width and im.height) and override the x and y. Also ensure correct orientation.
			

			var imwidth = parseInt(im.width());
			var optswidth = parseInt(opts.width);
			var imheight = parseInt(im.height());
			var optsheight = parseInt(opts.height);
			var outx = (opts.x * imwidth) / optswidth;
			var outy = (opts.y * imheight) / optsheight;
			
			opts.x = parseInt(outy);			//Note x/y swapped around on purpose
			opts.y = parseInt(outx);
			
			if(verbose == true) console.log("Pixel coords in image space = " + opts.x + "," + opts.y); 
			
		}
		
		
		
		im.convertHSVscale();	
		pixel = [];
		
		
		pixel = im.pixel(opts.x,opts.y);
		
		//Pixel is now the correct array e.g. [104,113,170] which is Hue,Saturation,Value (0-179, 0-255, 0-255)
		if(verbose == true) console.log('One pixel:' + JSON.stringify(pixel)); 
		
		
		
		
		//Write this array into the range calculation in the config file
		var readConfigFile = opts.photo.replace(".jpg", ".json");
		
		if ((opts.master == "false")||(opts.master == false)) {
				 //Yes, this photo already has a json file to use
				 console.log("Using existing photo-specific config file.");
		} else {
			//Otherwise, use the master file by default
			console.log("Using master config file.");
			readConfigFile = masterConfigFile;
			
			//Note: possible improvement here - we can visibly affect the current photo also with this new colour.
		}
		
		//Read in the global config to see if we should be using a shared drive version of the config
		readConfig(globalConfigFile, function(globConf, err) {
			if(err) {
				//There was a problem loading the global config
				console.log("Error reading global config file:" + err);
				
				var resp = {
							"completed": false,
							"error": "Error reading global config file"	   	 
				} 
				console.log("returnParams:?CUSTOMJSON=" + encodeURIComponent(JSON.stringify(resp)));

			
			} else {
			
				//With this global config, determine whether we should use a remote version
				if(globConf.backupTo && globConf.backupTo[0]) {
					//By default use the destination version, since that is shared. It sits in the home folder of the shared drive folder						
					
					var sharedMasterConfigFile = globConf.backupTo[0] + "/master.json"; 
					try {
					  if (fs.existsSync(sharedMasterConfigFile)) {
						//master file exists
						//And copy shared master down to local master
						try {
						  fsExtra.copySync(sharedMasterConfigFile, masterConfigFile);
						  console.log('Copied shared master ' + sharedMasterConfigFile + ' to local master ' + masterConfigFile + ', success!');
						} catch (err) {
						  console.error(err);
						}
						
						
						
					  } else {
						//doesn't exist, so local master will be backed up, also
						try {
						  fsExtra.copySync(masterConfigFile, sharedMasterConfigFile);
						  console.log('Shared local master config, success!');
						} catch (err) {
						  console.error(err);
						}
					  }
					  
					  //Either way we need to read the shared version herre
					  if ((opts.master == "false")||(opts.master == false)) {
					  		//Ignore variations here
					  } else {
					  	readConfigFile = sharedMasterConfigFile;
					  }
					  
					} catch(err) {
					  console.error(err);
					}
					
					
					
				} else {
					//Otherwise use the local version
				}
		
				console.log("Using config file: " + readConfigFile);
		

				//Start of reading file config file
				readConfig(readConfigFile, function(conf, err) {
				
					   var oldNonColourFactors = null;
				
					   if(err) {
						 console.log("Error reading config file:" + err);
						 process.exit(0);
					   }
			   
			   
					   if(opts.type == "wound") {
							var colourRanges = conf.input.wound.colourRanges;
						} else {
							var colourRanges = conf.input.sticker.colourRanges;
						}
			   
					   if(opts.keep == "false") {
							//We want to remove what is already there
							if(colourRanges[0]) {
								oldNonColourFactors = {
									"dilationIterations": colourRanges[0].dilationIterations,
									"gaussianBlur": colourRanges[0].gaussianBlur,
									"edgeDetectLower": colourRanges[0].edgeDetectLower,
									"edgeDetectUpper": colourRanges[0].edgeDetectUpper
								};
							}
							colourRanges.length = 0;
					   }
			   
					   if(opts.lowRange == "true") {
			   
							//Create a new entry in colour ranges
							var len = colourRanges.length;
							if(colourRanges[0]) {
								var newRange = {
									"lowerHSV": pixel,
									"upperHSV": pixel,
									"dilationIterations": colourRanges[0].dilationIterations,
									"gaussianBlur": colourRanges[0].gaussianBlur,
									"edgeDetectLower": colourRanges[0].edgeDetectLower,
									"edgeDetectUpper": colourRanges[0].edgeDetectUpper,
									"comment": "Created by hue-due.js. Run again for upper."		
								};
						
						
							} else {
								//Keep the old dilation factors etc.
								if(oldNonColourFactors) {
									var newRange = {
										"lowerHSV": pixel,
										"upperHSV": pixel,
										"dilationIterations": oldNonColourFactors.dilationIterations,
										"gaussianBlur": oldNonColourFactors.gaussianBlur,
										"edgeDetectLower": oldNonColourFactors.edgeDetectLower,
										"edgeDetectUpper": oldNonColourFactors.edgeDetectUpper,
										"comment": "Created by hue-due.js. Run again for upper."		
									};
								} else {
									//Defaults
									var newRange = {
										"lowerHSV": pixel,
										"upperHSV": pixel,
										"dilationIterations": 4,
										"gaussianBlur": [7,7],
										"edgeDetectLower": 0,
										"edgeDetectUpper": 50,
										"comment": "Created by hue-due.js. Run again for upper."		
									};
						
								}
					
							}
					
							if(checkRangeExist(colourRanges, newRange) == false) {
									colourRanges[len] = newRange;
							}
					
					   } 
			   
			   
					   if(opts.lowRange == "false") {
							//Upper range, so set the last colourRange element with this upper value
							var len = colourRanges.length;
					
							var oldHSV = colourRanges[len - 1].lowerHSV;
							var newHSV = pixel;
					
							var lowerHSV = [];
							var upperHSV = [];
					
							//OK, get the lowest and the highest of the individual hsv components, and store them in that order
							for(dim = 0; dim < 3; dim++) {
								if(oldHSV[dim] < newHSV[dim]) {
									lowerHSV[dim] = oldHSV[dim];
									upperHSV[dim] = newHSV[dim];
								} else {
									lowerHSV[dim] = newHSV[dim];
									upperHSV[dim] = oldHSV[dim];
								}
							}
					
					
							colourRanges[len - 1].lowerHSV = lowerHSV;
							colourRanges[len - 1].upperHSV = upperHSV;
							colourRanges[len - 1].comment = "Created by hue-due.js";
			   
					   }
			   
					   //Or dark to light
					   if(opts.lowRange == "darkToLight") {
							//So, take this colour, get a slight delta on the hue up and down, and get a wide variation
							//of colour range from dark to light in terms of brightness and saturation
					
							//Pixel is in the correct array e.g. [104,113,170] which is Hue,Saturation,Value (0-179, 0-255, 0-255) 
							//Weird issue - it seems like 250 is the max, not 255.
							var lowH = pixel[0] - AUTO_HUE_RANGE;
							if(lowH < 0) lowH = 180 + lowH;		//wrap around
							var lowS = pixel[1] - AUTO_SATURATION_RANGE;
							if(lowS < 0) lowS = 0;
							var lowV = pixel[2] - AUTO_VALUE_RANGE;
							if(lowV < 0) lowV = 0;			   		
							var lowHSV = [lowH, lowS, lowV];		//Get full range array
					
							var highH = pixel[0] + AUTO_HUE_RANGE;
							if(highH > 180) highH = highH - 180;		//wrap around
							var highS = pixel[1] + AUTO_SATURATION_RANGE;
							if(highS > 250) highS = 250;
							var highV = pixel[2] + AUTO_VALUE_RANGE;
							if(highV > 250) highV = 250;			   		
							var highHSV = [highH, highS, highV];		//Get full range array
					
					
					
					
					
							var len = colourRanges.length;
							if(colourRanges[0]) {
								var newRange = {
									"lowerHSV": lowHSV,
									"upperHSV": highHSV,
									"dilationIterations": colourRanges[0].dilationIterations,
									"gaussianBlur": colourRanges[0].gaussianBlur,
									"edgeDetectLower": colourRanges[0].edgeDetectLower,
									"edgeDetectUpper": colourRanges[0].edgeDetectUpper,
									"comment": "Created by hue-due.js. Run again for upper."		
								};
							} else {
								//Keep the old dilation factors etc.
								if(oldNonColourFactors) {
									var newRange = {
										"lowerHSV": lowHSV,
										"upperHSV": highHSV,
										"dilationIterations": oldNonColourFactors.dilationIterations,
										"gaussianBlur": oldNonColourFactors.gaussianBlur,
										"edgeDetectLower": oldNonColourFactors.edgeDetectLower,
										"edgeDetectUpper": oldNonColourFactors.edgeDetectUpper,
										"comment": "Created by hue-due.js. Run again for upper."		
									};
								} else {
									//Defaults
									var newRange = {
										"lowerHSV": lowHSV,
										"upperHSV": highHSV,
										"dilationIterations": 4,
										"gaussianBlur": [7,7],
										"edgeDetectLower": 0,
										"edgeDetectUpper": 50,
										"comment": "Created by hue-due.js. Run again for upper."		
									};
						
								}
					
							}
					
							if(checkRangeExist(colourRanges, newRange) == false) {
									colourRanges[len] = newRange;
							}
					   }
			   
			   
			   
			   
			   
			   
			   
					   //And write out the config file again
					   writeConfig(readConfigFile, conf, function(err) {
							if(err) {
								console.log("Error writing .json file:" + err);
							} else {
								//And write it out to the network version if that exists. Important to do before
								//running show-adjust-colours.js below, which will use the network version.
								if(globConf.backupTo && globConf.backupTo[0]) {
									//By default use the destination version, since that is shared. It sits in the home folder of the shared drive folder						
					
									var sharedFile = globConf.backupTo[0] + '/' + origPhoto.replace(".jpg", ".json");
									try {
										console.log("Saving config to network drive: " + sharedFile);
										fs.writeFileSync(sharedFile, JSON.stringify(conf, null, 6));
									} catch (err) {
									   console.error(err);
									}
								}
								
								
								
								console.log(".json file updated successfully with new colour range!");
					
								//Run a new process to display the new page			
								var run = 'node ' + __dirname + '/show-adjust-colours.js ' + process.argv[2];
								exec(run, function(error, stdout, stderr){
									if(error) {
										console.log("Error starting MedImage:" + error);
								
									} else {
										//Success - create a backup
			  							 if(readConfigFile == sharedMasterConfigFile) {
											//Want to do the backup of this file ourselves - because it isn't in the normal
											//folders. 
											//Copy global master down to local master
											try {
											  fsExtra.copySync(sharedMasterConfigFile, masterConfigFile);
											  console.log('Backing up global master ' + sharedMasterConfigFile  + ' to local ' + masterConfigFile + ' success!');
											} catch (err) {
											  console.error(err);
											}
											//No need for a backupFiles signal to the main parent MedImage
										 } else {
											if(readConfigFile == masterConfigFile) {
												//Save it up as the global master
												try {
												  fsExtra.copySync(masterConfigFile, sharedMasterConfigFile);
												  console.log('Backing up master ' + masterConfigFile  + '  to global master ' + sharedMasterConfigFile + ' success!');
												} catch (err) {
												  console.error(err);
												}
											} else {
											
												var photoFile = normalizeInclWinNetworks(resolve(readConfigFile)).toString();
											
												console.log("backupFiles:" + photoFile);  	//Must be relative
											}
										
											
										}
								
										console.log(stdout);
								
									}
								});
					
						
						
						
	 
							}
					   });
			   
				});			//End of reading file config file
			}  //end of else reading global config
		});  //end of read global config

		
	});
			

	



} else {
	var example = {
		"photo": "/a/path/here.jpg",
		"x": 345,
		"y": 678,
		"lowRange": true,
		"type": "wound",
		"master": true
	};

	console.log("Usage: node hue-due.js file/path/run.json where run.json is something like:\n " + JSON.stringify(example, null, 2));


}