/*
MedImage Wound Analysis Add-on
==============================

Developed by (C) AtomJump Ltd. 9 Aug 2017. This is a commercial add-on to MedImage, and is licensed
by AtomJump on a per-server basis.


This script is run from the MedImage server, or independently as
node wound-size.js photo/path/file.jpg

and this photo should include a sticker and a wound. Preferably do not include background colours
that match the sticker colour in the photo, as this could result in a non-detection of the sticker
and a resulting mismatch of the area of the wound. The input path is relative to the config/master.json
photosRootPath parameter.

It will output 
photo/path/file.json
and
photo/path/file.wound-view.jpg

The .json is a descriptive file with the results for that photo, including the size in square cm of the wound.
The wound itself may be ambiguous between several different parts of the photo. These will be listed as 
ambiguous options in the json, and by default, the closest to the centre of the photo is used.  
However, the .json can be modified to include a different centre by modifying input.selectedWoundPoint, and 
then this script is run again.

The .wound-view.jpg file, is a smaller version of the photo, with the highlighted sticker and wound. Other
ambiguous options for the wound are also displayed in a different colour. This allows a different script to 
allow user feedback on which wound is the correct one, by selecting a point in this smaller photo, and then
modifying input.selectedWoundPoint (which will be in the original scale).

Typically, all the colour ranges and other parameters will initially come from the config/master.json file,
but they can be modified by hand on an individual photo basis for finer control.



Installation
============

Firstly, you will need nodejs and opencv installed.

For opencv, as mentioned at http://milq.github.io/install-opencv-ubuntu-debian/

sudo apt-get install libopencv-dev python-opencv
will work on Ubuntu, and this is adequate. That doc also mentions the offical route.

Then, as mentioned by https://www.npmjs.com/package/opencv

npm install opencv
npm install upath

However, Windows is the exception. You will need a valid license (the free community edition is potentially possible)
for Visual Studio. This package costs around US$500 for a professional version.




Research
========

https://stackoverflow.com/questions/29156091/opencv-edge-border-detection-based-on-color

See
https://community.risingstack.com/opencv-tutorial-computer-vision-with-node-js/
for a tutorial on NodeJS OpenCV

and
https://github.com/peterbraden/node-opencv/wiki/Matrix
for commands in NodeJS

For an HSV colour editor see:
http://colorizer.org

First attempts:

rgb of blister
185 105 97
= 5 38 55  in HSV (0-360, 0-100, 0-100)

Opencv uses HSV 0-180, 0-255, 0-255
So blister = Open CV HSV 2 96 140 

rgb of ordinary skin
200 150 139
= 10 35 66  in HSV (0-360, 0-100, 0-100)
So skin = Open CV HSV 5 89 168

Also see https://www.researchgate.net/publication/259644745_Segmentation_of_Chronic_Wound_Areas_by_Clustering_Techniques_Using_Selected_Color_Space


Wound colour ranges

s = 37 - 100
v = 14 - 60
	
in OpenCV HSV: s = 94 - 255
	           v = 35 - 154
	
Sticker, or an NZ 50c coin:
radius of 50c coin 2.4 cm. Area = 4.52389 square cm


Search google for 'area of circle', for quick sizing calculator. Use more details for 4dp.


*/


var cv = require('opencv');
var fs = require('fs');
var upath = require('upath');
var path = require('path');
var queryString = require('querystring');
var resolve = require('path').resolve;
var fsExtra = require('fs-extra');
var area = require('area-polygon');


var masterConfigFile = __dirname + '/config/master.json';
var mainMedImagePath = "../../photos/";
var mainMedImagePathAfterBackup = "../../photos/";
var globalConfigFile = __dirname + '/../../config.json';
global.mainMedImagePathAfterBackup = mainMedImagePathAfterBackup;

var verbose = false;			//Should be 'false' before a release
var webForm = false;		//Whether we came from a web form, or on the command line



//Include shared funcs
var normalizeInclWinNetworks = require(__dirname + '/common/libfuncs.js').normalizeInclWinNetworks;
var readConfig = require(__dirname + '/common/libfuncs.js').readConfig;
var writeConfig = require(__dirname + '/common/libfuncs.js').writeConfig;

//Include area functions
var refresh = require(__dirname + '/common/refresh-thumbnail.js'); 


function getUserDrawings(conf, img, width, height)
{
	//Returns contours object based off an array of points drawn by a human with a mouse or on a tablet.
	//The coordinates are in an array format ie.  [[x1,y1],[x2,y2],[x3,y3] ... ], and are already sitting in
	//the config file conf, in the correct pixel scales
	if(verbose == true) console.log("Getting user drawings");
	var WHITE = [255];	
	var BLACK = [ 0];
	var lineWidth = 8;
	var newSketches = [];
	
	if(conf.input.userSketch) {
	
		//Create a buffer matrix
		var bufferMat = img.copy();
		bufferMat.convertGrayscale();
		bufferMat.rectangle([0,0],[width,height], BLACK, -1);	
		
		if(verbose == true) console.log("Drawings exist, created buff");
		
		var sketch = conf.input.userSketch;
		
		for(var cnt=0; cnt< sketch.length; cnt++) {
			//Loop through all of the points	
			
			if(verbose == true) console.log("Using drawing " + cnt);
			if(sketch[cnt].newSketch == true) {
				sketch[cnt].newSketch = false;		//Switch off any chance of this being considered new next time
				newSketches.push(cnt);		//It is likely that this id number for this area, should be selected by default			
											//Note: this is not precise, and if there are overlapping user drawings
											//it may not be an accurate reflection. Worst case, it isn't selected by default
			}	
			
			bufferMat.fillPoly([sketch[cnt].points], WHITE); 
		
		}
		
		//OK drawn all the lines onto the buffer. Now generate a contour object
		var contours = bufferMat.findContours();
		delete bufferMat;
		
		return { "contours": contours, "newSketches": newSketches };
	
	}
	return false;
}



function getZoomBoundaryFromCoords(x, y, screenWidth, screenHeight, photoWidth, photoHeight)
{
	//Input x,y in screen coords. screenWidth = 800 x 600 typically. photoWidth = 3000, 2000 approx
	//We want to output [ left, right, top, bottom ] in photo coords of the boundary of the photo around the x,y,
	//but the output screen dimensions are kept the same.
	
	
	var zoomScale = 0.4;				//Ratio of photo pixels to new zoomed pixels
	var zoomWidth = parseInt(photoWidth * zoomScale);	
	var zoomHeight = parseInt(photoHeight * zoomScale);
	
	var outx = (x * photoWidth) / screenWidth;
	var outy = (y * photoHeight) / screenHeight;
					
	var photoX = parseInt(outx);			//Note x/y were swapped around on purpose
	var photoY = parseInt(outy);

	var left = photoX - (zoomWidth / 2);
	if(left < 0) left = 0;			//border with the photo itself
	
	var right = left + zoomWidth;
	if(right > photoWidth) {
		right = photoWidth;
	}

	var top = photoY - (zoomHeight / 2);
	if(top < 0) top = 0;
	
	var bottom = top + zoomHeight;
	if(bottom > photoHeight) bottom = photoHeight;
	
	var retObj = {
		"left": left,
		"right": right,
		"top" : top,
		"bottom" : bottom	
	}
	return retObj;


}





function isEncoded(uri) {
  uri = uri || '';

  return uri !== decodeURIComponent(uri);
}


function convertCoords(x, y, optswidth, optsheight, imwidth, imheight, zoomedIn, zc) {
	//Inputs coords in screen-view coords (i.e. (0 - 800) x (0 - 600) typically), and convert these
	//into coordinates of the source image (e.g. (0 - 3000ish) x (0 - 2000ish).
	//If the zoomedIn option is set to true, it means we are using a zoomed in version of the photo,
	//zoomedInCoords (zc) will be [left, right, top, bottom] in coordinates from the source image.
	
	//Coords are 0,0 top left of the image.

	if(zoomedIn == false) {
		var outx = (x * imwidth) / optswidth;
		var outy = (y * imheight) / optsheight;
					
		var newX = parseInt(outx);			//Note x/y were swapped around on purpose
		var newY = parseInt(outy);
		
		var retObj = {
			"newX": newX,
			"newY": newY	
		};
	} else {
		//Zoomed in version
		var outx = (((zc.right - zc.left) * x) / optswidth) + zc.left;
		var outy = (((zc.bottom - zc.top) * y) / optsheight) + zc.top;
		
		var newX = parseInt(outx);			//Note x/y were swapped around on purpose
		var newY = parseInt(outy);
		
		var retObj = {
			"newX": newX,
			"newY": newY	
		};
	}

	return retObj;
}





function trailSlash(str)
{
	if(str.slice(-1) == "/") {
		return str;
	} else {
		var retStr = str + "/";
		return retStr;
	}

}

function capitalise(photo, medImagePath, cb) {
	//Input: photo = full photo path
	//       medImagePath = folder that contains the photos relative to 
	// 
	
	//Auto capitalise the first directory path - which means changing the default ID dir (the 
	//entry before the 
	//			test/path/photos/nhId/photo.jpg
	//becomes
	//  		test/path/photos/NHID/photo.jpg
	//But note that the root folder photos should not be affected:
	//          test/path/photos/photo.jpg
	//remains the same (and does NOT become test/path/PHOTOS/photo.jpg)
	//Returns cb(photoFileOnly, newDirPath, rootPath) 
	
	console.log("Capitalising " + photo + " at " + medImagePath);
	
	var thisCb = cb;
	
	readConfig(globalConfigFile, function(conf, err) {
		if(err) {
			//There was a problem loading the global config
			console.log("Error reading global config file:" + err);
			thisCb(photo, null, null);
		} else {
			
			if(verbose == true) console.log("Read global config OK.");
			if (typeof thisCb === 'undefined') {
				console.log("Error: callback was not defined.");
				thisCb(photo, null, null);
			
			}
			
			
			var rootPath = normalizeInclWinNetworks(medImagePath);
			if(verbose == true) console.log("Root path:" + medImagePath);
			
			
			var normalizedFile = normalizeInclWinNetworks(photo);
			normalizedFile = normalizedFile.replace(rootPath, "");
			if(normalizedFile[0] == '/') {
				normalizedFile = normalizedFile.substr(1);	//Remove any '/'s at the start
			}
			if(verbose == true) console.log("Photo file (Normalized):" + normalizedFile);		
			
			
			var pathRenaming = [];
			pathRenaming[0] = rootPath;
			var pathSections = [];
			var filenameSections = [];
			var cnt = 0;
	
			var rootPath = pathRenaming[cnt];
		
			if(verbose == true) console.log("Using rootPath:" + rootPath);
			
			//Append the normalized file to this path
			var fullFile = rootPath + normalizedFile;
			
			if(verbose == true) console.log("About to split:" + fullFile);
			
			filenameSections[cnt] = normalizedFile.split("/");
			pathSections[cnt] = fullFile.split("/");
			if((filenameSections[cnt].length > 1) && (filenameSections[cnt][0]) && (filenameSections[cnt][0] != '')) {
				//We have an NHID.
				
				var photoFilename = pathSections[cnt].pop();		//Remove the file itself
				
				console.log(JSON.stringify(filenameSections[cnt]));
				console.log(JSON.stringify(pathSections[cnt]));
				var oldPath = pathSections[cnt].join("/");

				if(verbose == true) console.log("oldPath: " + oldPath);


				//Make all the sections in the filename upper case
				var fnSections = filenameSections[cnt].length - 1;		//The minus 1 doesn't include the trailing filename itself
				var ptSections = pathSections[cnt].length - fnSections;		//Get the difference for the 'path only' sections
				for(var section = 0; section < fnSections; section ++) {
					pathSections[cnt][ptSections + section] = pathSections[cnt][ptSections + section].toUpperCase();
				}

				var newPath = pathSections[cnt].join("/");
				
				if(verbose == true) console.log("newPath: " + newPath);
				
				var oldDirPath = oldPath;
				var newDirPath = newPath;
				console.log("Renaming path " + oldDirPath + " to " + newDirPath);
	
				if(oldDirPath != newDirPath) {
					//Wait until photo file is in the old path fully
		
					var platform = process.platform;
					var isWin = /^win/.test(platform);
					if(isWin) {
						//Ensure the new folder is made
						fsExtra.mkdirsSync(newDirPath);
						
						
						//If this is a Windows shared drive, we want to change into the folder first.
						var winSharedDrive = false;
						if((oldDirPath[1]) && (oldDirPath[1] == ':') && (oldDirPath[0].toUpperCase() != 'C')) {
							winSharedDrive = true;
							var oldCwd = process.cwd();
							//Very Likely a Windows path. 
							//Change the drive folder.
				
							process.chdir(oldDirPath.charAt(0).toUpperCase() + ":/");
							
						}
						
						
						
						//We only need to rename the directory	
						fs.rename(oldDirPath, newDirPath, function(err) {
							if(err) {
								console.log("Error renaming path: " + err);
								
								//If the folder is already made, return the new path anyway, else return the old path
								fs.lstat(newDirPath, function(err, stats) {
									if(!err) {
										//OK the 
										console.log("Checking if folder exists status:" + stats.isDirectory());
										if(stats.isDirectory()) {
											thisCb(newDirPath + '/' + photoFilename, newDirPath, rootPath);
										} else {
											//OK, some other issue - use the old folder
											thisCb(oldDirPath + '/' + photoFilename, oldDirPath, rootPath);	
										}
									} else {
										//An error, use the old non capitalised folder
										console.log("Checking if folder exists status:" + stats.isDirectory());
										if(stats.isDirectory()) {
											thisCb(newDirPath + '/' + photoFilename, newDirPath, rootPath);
										} else {
											//OK, some other issue - use the old folder
											thisCb(oldDirPath + '/' + photoFilename, oldDirPath, rootPath);	
										}
									
									}
								
								});
								
									
								
							} else {
								
								if(verbose == true) console.log("Calling back from renaming.");										
								thisCb(newDirPath + '/' + photoFilename, newDirPath, rootPath);	
								
							}
							
							if(winSharedDrive == true) {
								//Change back to the old drive anyway
								process.chdir(oldCwd);
							}
							
						});
						
					} else {
						//This was the way we did it on unix (tested OK) - this may have preserved all old
						//content which was at the old name? But it seems like overkill. Would a
						//simple rename suffice?
					
					
						//Ensure the new folder is made
						fsExtra.mkdirsSync(newDirPath);
		
						
						//Copy all contents into the new folder (only overwrite if already there)
						fsExtra.copy(oldDirPath, newDirPath, { overwrite: true, preserveTimestamps: true }, function(err) {
							if(err) {
								console.log("Error copying older path to new path: " + err);
								
								if(verbose == true) console.log("Removed old folder cnt=" + cnt);	

								//If the folder is already made, return the new path anyway, else return the old path
								fs.lstat(newDirPath, function(err, stats) {
									if(!err) {
										//OK the 
										console.log("Checking if folder exists status:" + stats.isDirectory());
										if(stats.isDirectory()) {
											thisCb(newDirPath + '/' + photoFilename, newDirPath, rootPath);
										} else {
											//OK, some other issue - use the old folder
											thisCb(oldDirPath + '/' + photoFilename, oldDirPath, rootPath);	
										}
									} else {
										//An error, use the old non capitalised folder
										console.log("Checking if folder exists status:" + stats.isDirectory());
										if(stats.isDirectory()) {
											thisCb(newDirPath + '/' + photoFilename, newDirPath, rootPath);
										} else {
											//OK, some other issue - use the old folder
											thisCb(oldDirPath + '/' + photoFilename, oldDirPath, rootPath);	
										}
									
									}
								
								});
										
								
							} else {
				
								//Then remove the old folder and contents - if on unix.						 
								var topLevelFolder = path.join(rootPath, filenameSections[cnt][0]);
								if(verbose == true) console.log("About to remove top-level folder " + topLevelFolder);		
								
								fsExtra.removeSync(topLevelFolder);		//Removes the whole tree off the 1st folder name. Equiv to 'rm -rf'
								if(verbose == true) console.log("Removed top-level folder " + topLevelFolder);		
									
								
								//If last entry i.e. the before-backup path is returned
								thisCb(newDirPath + '/' + photoFilename, newDirPath, rootPath);	
								
							}
						});
					}	
		
				
					
				} else {
					//OK - it was the same path, but we still need to return
					thisCb(photo, null, rootPath);	
										
				
				}	 //end of 	oldDirPath == newDirPath
			}		//end of valid filename sections
		} //end of successful read config	
		
	});	//end of read config

}





  
function insertSticker(conf, areaSquarePixels, points, woundArea)
{

	

	var area = {
			"type": "sticker",
			"status": "add",
			"handDrawing": false,
			"hideMidpoints": false,
			"areaSquarePixels": areaSquarePixels,
			"areaSquareCm": Number(conf.input.sticker.physicalAreaSquareCm),
			"midPointPixels": {
				"x": 0,
				"y": 0
			},
			"points": points
		};
	
	if(!conf.output.areas) conf.output.areas = [];
	
	//Remove any existing sticker areas
	for(var cnt = 0; cnt< conf.output.areas.length; cnt++) {
		console.log("Type is " + conf.output.areas[cnt].type);
		if(conf.output.areas[cnt].type === "sticker") {
			conf.output.areas.splice(cnt);
			console.log("Removed sticker area " + cnt);
		}
	}
	
	
	conf.output.areas.push(area);
	conf.output.stickerSquarePixels = areaSquarePixels;
	if(!conf.output.fullWound.areaSquareCm) {
		conf.output.fullWound.areaSquareCm = woundArea;
	}
	
	conf = refresh.calculateDrawingArea(conf);

	
	return conf;


}  
  

function detectSticker(conf, im)
{
	//Image process and try to auto-determine the sticker boundary as a circular shape
	
	//Loop through all the colour ranges and try to find the largest circular shape
	//that represents the sticker.
	var largestI = null;
	var largestArea = 0;
	var largestContours = {};


	for(var cnt = 0; cnt < conf.input.sticker.colourRanges.length; cnt++) {

	  var im_canny = im.copy();
	  im_canny.convertHSVscale();	

	  var rng = conf.input.sticker.colourRanges[cnt];

	  if(verbose == true) console.log("Checking from range:" + JSON.stringify(rng.lowerHSV) + " to " + JSON.stringify(rng.upperHSV));

	  im_canny.inRange(rng.lowerHSV, 
						rng.upperHSV);

	  if(verbose == true) console.log("Blur: " +JSON.stringify(rng.gaussianBlur));

	  im_canny.gaussianBlur(rng.gaussianBlur);		//Perhaps don't need this blur

	  if(verbose == true) console.log("Edge detect from:" + JSON.stringify(rng.edgeDetectLower) + " to " + JSON.stringify(rng.edgeDetectUpper));

	  im_canny.canny(rng.edgeDetectLower, 
					rng.edgeDetectUpper);

	  if(verbose == true) console.log("Dilation: " + rng.dilationIterations);
	  im_canny.dilate(rng.dilationIterations);

	  var contours = im_canny.findContours();



	  for(i = 0; i < contours.size(); i++) {
		

		if(contours.area(i) > conf.input.sticker.minAreaSquarePixels) {
			var arcLength = contours.arcLength(i, true);
			contours.approxPolyDP(i, 0.01 * arcLength, true);

			if((contours.cornerCount(i) > 5) && (contours.cornerCount(i) < 20)) {
				 //Likely to be a circle between 9 - 16
				console.log("Area:" + contours.area(i));
				if(verbose == true) console.log("Corner count:" + contours.cornerCount(i));         

				if(contours.area(i) > largestArea) {
				  largestArea = contours.area(i);
				  largestI = i;	
				  largestContours = contours;
				}
			}

		}  //end of 
	  }  //end of sticker contours
	}  //end of stricker colour ranges


	if(largestI) {
	   //There appears to be a circular shape in the picture
	   if(verbose == true) console.log("Sticker is " + largestArea + " pixels squared");
	   conf.output.stickerSquarePixels = largestArea;
	   
	   //Now add the contours so that the edge can be edited.
	   // Access vertex data of contours
	   
	   var points = [];
	   
		for(var c = 0; c < largestContours.size(); ++c) {
		  console.log("Contour " + c);
		  if(c == largestI) {
		  
			  for(var i = 0; i < largestContours.cornerCount(c); ++i) {
			
					var point = largestContours.point(c, i);
					console.log("(" + point.x + "," + point.y + ")");
					var thisPoint = [ point.x, (conf.output.scaledPhoto.height - point.y) ];		//Was getting coords upside down
					points.push(thisPoint);	
			
			  }
		  }
		}
	   
	   conf = insertSticker(conf, largestArea, points, 0);
	   
	} else {
		if(verbose == true) console.log("Sorry, no sticker was detected.");
		//Do not insert a sticker in here.
	
	}

	return conf;
	

}

function calculateMidpoint(config, areaId)
{
	//Finds the midpoint of a particular area
	var points = config.output.woundAreas[areaId].points;

	var xmid = 0;
	var ymid = 0;

	for(var cnt = 0; cnt< points.length; cnt++) {
		xmid += points[0];
		ymid += points[1];
	}
	
	config.output.woundAreas[areaId].midPointPixels = {
		"x": Math.round(xmid / points.length),
		"y": Math.round(ymid / points.length)
	}
	
	return config;
	
}



function getBackupFoldersCapitalised(inputPhoto, back) {
	//Input an array of input photo names,
	//Output capitalised versions of those photo file names, with the output folder correctly created if not done
	//already.
	
	//Returns cb(renamedFolder)
	
  	//Capitalise backup folders 
  	readConfig(globalConfigFile, function(globalConf, err) {
		if(err) {
			//There was a problem loading the global config
			console.log("Error reading global config file:" + err);
			back("Error reading global config file:" + err);
		} else {

			if(verbose == true) console.log("Read global config OK.");
			
			if(globalConf.backupTo && globalConf.backupTo[0]) {
				var mainMedImagePathAfterBackup = trailSlash(globalConf.backupTo[0]);	  //Relative to __dirname? or process.cwd()?
			} else {
				//Leave as-is, global
				var mainMedImagePathAfterBackup = global.mainMedImagePathAfterBackup;
			}
			
			if(verbose == true) console.log("mainMedImagePath after backup: " + mainMedImagePathAfterBackup);
			
			var photoFilename = mainMedImagePathAfterBackup + inputPhoto;
			
			//Now capitalise on the global backup folder
			capitalise(photoFilename, mainMedImagePathAfterBackup, function(photoFileOnly, newDirPath, rootPath) {
				 //Do nothing here - this is prepping the way for the correct folder name in the target backup dirs.
				console.log("Finished capitalising target folder.");
				back(null);
			});
								
		}
	});
	
	return;

}



function processPhoto(photoFile, opts, cb) {

  console.log("Reading photo file:" + photoFile);

  var readConfigFile = photoFile.replace(".jpg", ".json");
  var writeConfigFile = readConfigFile;		//we still want to write to this one, even if we don't read from it

  if (fs.existsSync(readConfigFile)) {
	 //Yes, this photo already has a json file to use
	 console.log("Using existing photo-specific config file.");
  } else {
  	  //OK check if a file off from the root folder exists
  	  readConfigFile = path.resolve(path.join(__dirname, mainMedImagePath, readConfigFile));
  	  if (fs.existsSync( readConfigFile)) {
  	  		console.log("Using existing absolute photo-specific config file: " + readConfigFile);
  	  		photoFile = path.resolve(path.join(__dirname, mainMedImagePath, photoFile));		//Also make absolute
  	  		writeConfigFile = readConfigFile;
  	  } else {
  	  			
			//Otherwise, use the master file by default
			console.log("Using master config file. " + readConfigFile + " not found.");


			readConfigFile = masterConfigFile;			//Change the read config file
				
	  }
  }


	readConfig(readConfigFile, function(conf, err) {

	   if(err) {
		 console.log("Error reading config file:" + err);
		 process.exit(0);
	   }
	   
	   
	   	if(verbose == true) console.log("About to read photo:" + photoFile);
		cv.readImage(photoFile, function(err, im) {
		  if (err) throw new Error("Error reading photo:" + photoFile + " Error:" + err);

		  conf.input.thisPhotoPath = photoFile;			//Set the current photo into the config
		  conf.input.scriptRootPath = __dirname + '/';

		  var width = im.width();
		  var height = im.height();
		  conf.input.photoSize.width = width;
		  conf.input.photoSize.height = height;

	
		 

		  if (width < 1 || height < 1) throw new Error("Sorry, the image " + photoFile + " needs to have a valid size. Width =" + width + " Height=" + height);

		  var outputPhoto = new cv.Matrix(height, width);
		  outputPhoto = im.copy();
	 
  
  
		  //Rescale original photo for quick web view of the original. No other drawings on this.
		  var webPhotoFile = photoFile.replace(conf.output.scaledPhoto.nameConvention.replaceText, 
												conf.output.scaledPhoto.nameConvention.withWebViewText);
		  var webPhoto = new cv.Matrix(height, width);
		  webPhoto = im.copy();		//Copy of the original
		  webPhoto.resize(conf.output.scaledPhoto.width, conf.output.scaledPhoto.height);
		  if(verbose == true) console.log('Saving web photo...' + webPhotoFile);
		  webPhoto.save(webPhotoFile);
		 
		  if(verbose == true) console.log('Done');


		   //Auto-detect the sticker off the smaller photo
		  conf = detectSticker(conf, webPhoto);
		  
		  //Now we're done with this image in memory
		  delete webPhoto;

		 
		  
		  refresh.refreshThumbnail(photoFile, webPhotoFile, conf, function(err, thumbnailPhotoFile) {
					
					
			  if(err) {
			  	   //And create a thumbnail anyway
				  var thumbnailPhoto = new cv.Matrix(height, width);
				  thumbnailPhoto = im.copy();		//Copy of the original
				  var thumbnailPhotoFile = photoFile.replace(conf.output.scaledPhoto.nameConvention.replaceText, 
														conf.output.scaledPhoto.nameConvention.withThumbnailText);

				  thumbnailPhoto.resize(conf.output.scaledPhoto.thumbnailWidth, conf.output.scaledPhoto.thumbnailHeight);
				  if(verbose == true) console.log('Saving thumbnail photo...' + thumbnailPhotoFile);
				  thumbnailPhoto.save(thumbnailPhotoFile);
				  delete thumbnailPhoto;
				  if(verbose == true) console.log('Done');
			  }


			  console.log('All images saved.');
  
  
  
  

			  //Now write back the config to the correct path
			  writeConfig(writeConfigFile, conf, function(err) {
				  if(err) {
						throw new Error("Cannot write photo config file " + writeConfigFile + "  err:" + err);
				  } else {
  
					  console.log("All finished! Written:" + writeConfigFile);

						
					   getBackupFoldersCapitalised(opts.originalPath, function(err) {
  
						   //Back up all the created/modified files to any secondary directories the user has specified in their MedImage server config
						   console.log("backupFiles:"  + resolve(webPhotoFile) + ";" + resolve(thumbnailPhotoFile) + ";" + resolve(writeConfigFile));
						   //Without the main photoFile (that gets backed up by default anyway): + resolve(photoFile) + ";"


						   //Check if we want to display the sticker version instead of the default wound version
						   var startWith = "wound";
						   if((conf) && (conf.output)) {
								if(!conf.output.stickerSquarePixels) {		//Either 0 or not existing
									startWith = "sticker";
								}
						   } else {
								startWith = "sticker";
						   }
   
						   if(opts.startWith == "sticker") {
								startWith = "sticker";
						   }



						  if(conf.output.fullWound) {
							  var outputArea = conf.output.fullWound.areaSquareCm;

						  } else {
							  var outputArea = "[Unknown Area]";
						  }
						  
						  if((conf) && (conf.output) && (conf.output.fullWound) && (conf.output.fullWound.underAreaSquareCm)) {
								
								var woundArea = conf.output.fullWound.woundAreaSquareCm;
								var underminingArea = conf.output.fullWound.underAreaSquareCm;
								var underminingAreaStr = ".<span style=\"color: #888;\"> External <i class=\"fa fa-pie-chart\"></i>  <span id=\"total-area\">" + woundArea + "</span> cm<sup>2</sup>.  Undermining <i class=\"fa fa-pie-chart\"></i>  <span id=\"total-area\">" + underminingArea + "</span> cm<sup>2</sup></span>";
						   } else {
								var underminingAreaStr = "";
						   }
						  
  
						  if(conf.output.areas) {
							  var areas = encodeURIComponent(JSON.stringify(conf.output.areas));
						  } else {
							 var areas = "[]";
						  }


						  var returnLink = "";
						  if(opts.ret) {
							 returnLink = opts.ret;
						  }

						  var sessionId = "";
						  if(opts.sessionId) {
							 sessionId = opts.sessionId;
						  }
  
						  var webRelativePhotoFile = opts.photo.replace(conf.output.scaledPhoto.nameConvention.replaceText, 
														conf.output.scaledPhoto.nameConvention.withWebViewText);
  
  
   						   //Display wound type classification
						   var startWoundType = "Unclassified";
						   if((conf) && (conf.output)) {
								if(conf.output.woundType) {		
									startWoundType = conf.output.woundType;
								} 
						   }
  

						  console.log("returnParams:?CUSTOMAREA=" + outputArea + "&CUSTOMIMAGE=" + opts.photo + "&CUSTOMWOUNDIMAGE=" + webRelativePhotoFile + "&AREAS=" + areas + "&UNDERMININGAREA=" + underminingAreaStr + "&CUSTOMSTARTWITH=" + startWith + "&CUSTOMWOUNDTYPE=" + startWoundType + "&RANDIMAGE=" + Date.now() + "&RETLINK=" + returnLink + "&SESSIONID=" + sessionId);
						  cb();
					  });		//End of getBackupFoldersCapitalised
					}
			   });	//End of writeConfig
				
		  });			//End of refresh thumbnail


		  cb();
		});  //end of cv.readImage

	    
	    
	   
	   
	});	//End of read config
	

}




 




//Starting from here.
var photoFile = "";


if(process.argv[2]) {
 
   var opts = {};
   
   if(isEncoded(process.argv[2])) {
  	 //Then this is from a web form.
  	 var opts = queryString.parse(decodeURIComponent(process.argv[2]));
  	 webForm = true;
  	 

  	 photoFile = normalizeInclWinNetworks(__dirname + "/" + mainMedImagePath + opts.photo);	
  	 
  	 if(verbose == true) console.log("Web operation photo file path converted to:" + photoFile);
  
  
  } else {
  
  	//Read the command line argument for the photo to process
  	photoFile = normalizeInclWinNetworks(process.argv[2]);		//This should be relative to the script folder
  	if(verbose == true)  console.log("Command line operation photo file path converted to:" + photoFile);
  	
  	opts.photo = photoFile;		//Trying this.
  	webForm = false;
  }
  
  opts.displayPath = opts.photo;
  
  //originalPath is always the short version of the photo file as it was entered eg. TEST/thisphoto.jpg or Test/thisphoto.jpg
  var srcFolder = normalizeInclWinNetworks(__dirname + "/" + mainMedImagePath);
  if(verbose == true)  console.log("Source folder path: " + opts.srcFolder);
  if(verbose == true)  console.log("Source photo: " + opts.photo);
  opts.originalPath = opts.photo.replace(srcFolder, "");
  console.log("Original path: " + opts.originalPath);
  
  
  
  readConfig(masterConfigFile, function(masterConf, err) {

	   if(err) {
	   	 console.log("Error reading config file:" + err);
	  	 process.exit(0);
	   }
	   
	   if(masterConf.output && masterConf.output.autoCapitaliseID && masterConf.output.autoCapitaliseID == true) { 
			 			 
			 var medImagePathBeforeBackup = normalizeInclWinNetworks(__dirname + "/" + mainMedImagePath);			 
			 
			 
			 			 
			 capitalise(photoFile, medImagePathBeforeBackup, function(photoFileOnly, newDirPath, rootPath) {
			 		
			 		//Finished capitalise
			 		console.log("Finished capitalise. Output photo:" + photoFileOnly + "  Output newDir:" + newDirPath);
			 		
			 		opts.displayPath = photoFileOnly.replace(rootPath, "");	
			 		
			 		
			 		
			 		var platform = process.platform;
					var isWin = /^win/.test(platform);
						
			 		
			 		
			 		//Rename the main one - photoFile
					if(webForm == true) {
						//We want the long version, but the new version
						if(newDirPath) {
						    if(photoFileOnly.search(newDirPath) >= 0) {
						    	//Already exists in there
						    	photoFile = photoFileOnly;
						    } else {
								photoFile = newDirPath + "/" + photoFileOnly;
							}
							
							
							if(isWin) {
								photoFile = photoFile.replace(rootPath, "");		//Not sure this replace is needed any more?						
								opts.photo = photoFileOnly.replace(rootPath, "");		//This should only be the display path.
							} else {
								opts.photo = photoFileOnly;
							}
						} else {
							//Which would be a complete photo path in this case anyway - leave as-is
						}
						console.log("Web operation photo file path changed to:" + photoFile);   //if(verbose == true) 
					} else {
						//We want the short version 
						if(isWin) {
							//On win we only want to update the opts.photo, but not the main folder passed in
							photoFileOnly = photoFileOnly.replace(rootPath, "");
						} else {
							//On linux, we only want the short version
							photoFile = photoFileOnly;	
						}
						opts.photo = photoFileOnly;		//This should only be the display path.	
						
						console.log("Command line photo file path changed to:" + photoFile);   //if(verbose == true) 
					}
					
				
					//Actually process the photos
					processPhoto(photoFile, opts, function() {});
					
			 });		//End of capitalise	
		} else {
				
			//Actually process the photos
			processPhoto(photoFile, opts, function() {});
  
  		}
  	});  //End of read master config


} else {

	console.log("Usage: node wound-size.js photo/path/file.jpg");
}



